﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;

  using Attributes;
  using Interpreter;
  using Stack;

  /// <summary>
  ///     Pushes the logical AND of the top two BOOLEANs.
  /// </summary>
  [PushExpression(StackTypes.Boolean, "BOOLEAN.AND")]
  public class BooleanAndExpression : PushResultExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack, 2, values => values[0] && values[1]);
    }
  }

  /// <summary>
  ///     Pushes the logical OR of the top two BOOLEANs.
  /// </summary>
  [PushExpression(StackTypes.Boolean, "BOOLEAN.OR")]
  public class BooleanOrExpression : PushResultExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack, 2, values => values[0] || values[1]);
    }
  }

  /// <summary>
  ///     Pushes the logical NOT of the top BOOLEAN.
  /// </summary>
  [PushExpression(StackTypes.Boolean, "BOOLEAN.NOT")]
  public class BooleanNotExpression : PushResultExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack, 1, values => !values[0]);
    }
  }

  /// <summary>
  ///     Pushes FALSE if the top FLOAT is 0.0, or TRUE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Boolean, "BOOLEAN.FROMFLOAT", StackTypes.Float)]
  public class BooleanFromFloatExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count == 0) return false;

      var value = Math.Abs(interpreter.FloatStack.Pop()) > 0.0000001;

      interpreter.BooleanStack.Push(value);
      return true;
    }
  }

  /// <summary>
  ///     Pushes FALSE if the top INTEGER is 0, or TRUE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Boolean, "BOOLEAN.FROMINTEGER", StackTypes.Integer)]
  public class BooleanFromIntegerExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count == 0) return false;

      var value = interpreter.IntegerStack.Pop() != 0;

      interpreter.BooleanStack.Push(value);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the logical AND of the top two BOOLEANs, whereby the first value is inverted first.
  /// </summary>
  [PushExpression(StackTypes.Boolean, "BOOLEAN.INVERT_FIRST_THEN_AND")]
  public class BooleanInvertFirstThenAnd : PushResultExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack, 2, values => !values[0] && values[1]);
    }
  }

  /// <summary>
  ///     Pushes the logical AND of the top two BOOLEANs, whereby the second value is inverted first.
  /// </summary>
  [PushExpression(StackTypes.Boolean, "BOOLEAN.INVERT_SECOND_THEN_AND")]
  public class BooleanInvertSecondThenAnd : PushResultExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack, 2, values => values[0] && !values[1]);
    }
  }


  /// <summary>
  ///     Pushes the logical AND of the top two BOOLEANs, whereby the second value is inverted first.
  /// </summary>
  [PushExpression(StackTypes.Boolean, "BOOLEAN.XOR")]
  public class BooleanXorExpression : PushResultExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack, 2, values => values[0] ^ values[1]);
    }
  }
}